@@ -16,6 +16,7 @@ class PipelineBuilder(object):
 
     :param optional pipeline: list of initial pipeline entries e.g. from another builder
     """
+
     def __init__(self, pipeline: list = None):
         self._pipeline = [] if pipeline is None else pipeline
         self._builder_id = _gen_id()
@@ -74,23 +75,28 @@ class PipelineBuilder(object):
         """
         args = self._unpack_single_tuple_args(args)
         task_type, parameters = self._task_type_and_parameters_for(fun)
-        self._pipeline.append(Task(_gen_id(), task_type, args, kwargs, parameters=parameters))
+        self._pipeline.append(Task.from_fields(_gen_id(), task_type, args, kwargs, **parameters))
         return self
 
-    def set(self, retry_policy:RetryPolicy = None, **params) -> 'PipelineBuilder':
+    def set(self, retry_policy: RetryPolicy = None, namespace: str = None,
+            worker_name: str = None) -> 'PipelineBuilder':
         """
         Sets task properties on the last entry added to the builder
 
         :param option retry_policy: the task retry policy to use
-        :param params: any other parameters to the task
+        :param option namespace: the task namespace
+        :param option worker_name: the task worker_name
         :return: the builder
         """
-        if retry_policy is not None:
-            params['retry_policy'] = retry_policy.to_proto()
 
         if any(self._pipeline):
             entry = self._pipeline[-1]
-            entry.set_parameters(params)
+            if retry_policy is not None:
+                entry.retry_policy.CopyFrom(retry_policy.to_proto())
+            if namespace is not None:
+                entry.namespace = namespace
+            if worker_name is not None:
+                entry.worker_name = worker_name
 
         return self
 
@@ -108,7 +114,7 @@ class PipelineBuilder(object):
         else:
             args = self._unpack_single_tuple_args(args)
             task_type, parameters = self._task_type_and_parameters_for(continuation)
-            self._pipeline.append(Task(_gen_id(), task_type, args, kwargs, parameters=parameters))
+            self._pipeline.append(Task.from_fields(_gen_id(), task_type, args, kwargs, **parameters))
         return self
 
     def get_destination(self):
@@ -129,32 +135,32 @@ class PipelineBuilder(object):
         """
         task_id = self._builder_id
         task_type = '__builtins.run_pipeline'
-        args = self.validate().to_proto(serialiser=serialiser)
+        args = self.validate().to_proto(serialiser)
         kwargs = {}
-        parameters = {}    
 
         # send a single argument by itself instead of wrapped inside a tuple
         if _is_tuple(args) and len(args) == 1:
             args = args[0]
 
-        task_request = TaskRequest(id=task_id, type=task_type, parameters=serialiser.to_proto(parameters))
-        serialiser.serialise_request(task_request, args, kwargs)
-        
+        task_request = TaskRequest(id=task_id, type=task_type)
+        args_and_kwargs = serialiser.serialise_args_and_kwargs(args, kwargs)
+        serialiser.serialise_request(task_request, args_and_kwargs)
+
         return task_request
 
     def finally_do(self, finally_action, *args, **kwargs) -> 'PipelineBuilder':
         """
         Adds finally to the pipeline
 
-        :param continuation: the python function which should be decorated with @tasks.bind()
+        :param finally_action: the python function which should be decorated with @tasks.bind()
         :param args: the task args
         :param kwargs: the task kwargs
         :return: the builder
         """
         args = self._unpack_single_tuple_args(args)
         task_type, parameters = self._task_type_and_parameters_for(finally_action)
-        task_entry = Task(_gen_id(), task_type, args, kwargs, parameters=parameters, is_finally=True)
-        task_entry.set_parameters({'is_fruitful': False})
+        combined_parameters = {**parameters, 'is_fruitful': False}
+        task_entry = Task.from_fields(_gen_id(), task_type, args, kwargs, is_finally=True, **combined_parameters)
         self._pipeline.append(task_entry)
         return self
 
@@ -176,15 +182,12 @@ class PipelineBuilder(object):
         """
         errors = []
 
-        for entry in self._pipeline:
-            entry.validate(errors)
-
         finally_tasks = [task for task in self._pipeline if isinstance(task, Task) and task.is_finally]
         if len(finally_tasks) > 1:
             errors.append('Cannot have more than one "finally_do" method')
         if len(finally_tasks) == 1 and finally_tasks[0] != self._pipeline[-1]:
             errors.append('"finally_do" must be called at the end of a pipeline')
-        
+
         if any(errors):
             error = ', '.join(errors)
             raise ValueError(f'Invalid pipeline: {error}')
@@ -201,21 +204,20 @@ class PipelineBuilder(object):
         return _Pipeline(self._pipeline, serialiser=serialiser).to_proto()
 
     @staticmethod
-    def from_proto(pipeline_proto: Pipeline, serialiser) -> 'PipelineBuilder':
+    def from_proto(pipeline_proto: Pipeline) -> 'PipelineBuilder':
         """
         Deserialises the pipeline from protobuf
 
         :param pipeline_proto: the pipeline as protobuf
-        :param serialiser: the serialiser to use such as DefaultSerialiser
         :return: Pipeline protobuf message
         """
         pipeline = []
 
         for proto in pipeline_proto.entries:
             if proto.HasField('task_entry'):
-                pipeline.append(Task.from_proto(proto, serialiser))
+                pipeline.append(Task.from_proto(proto))
             elif proto.HasField('group_entry'):
-                pipeline.append(Group.from_proto(proto, serialiser))
+                pipeline.append(Group.from_proto(proto))
 
         return PipelineBuilder(pipeline)
 
